// Program to send and receive periodic UDP heartbeats
// Bhaskaran Raman, bhaskar@cs.berkeley.edu, Nov 14 2000

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <assert.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>
#define PORT 2000
#define ALARM_TIME 5

struct hb_pkt {
  long seqno;
  char dummy[1424];
};
typedef struct hb_pkt hb_pkt_t;

/* The time at which heartbeats are started */
struct timeval start;
/* The socket descriptor */
int sock_fd;
/* Address on which to listen */
struct sockaddr_in listen_addr;
/* Port on which to listen */
int listen_port;
/* Address to which to send packets */
struct sockaddr_in to_addr;
/* Node to which to send packets */
char to_host[256];
/* Node to which to send packets */
struct in_addr to_ip;
/* Port to which to send packets */
int to_port;
/* The heart-beat interval (in ms) */
double hbInt;

/** To write output */
char foutStr[256];
FILE *fout;

struct file_entry {
  long seqno_recd;
  double time_recd;
};
typedef struct file_entry file_entry_t;

/* Return the current time in ms as double */
inline double now() {
  struct timeval t;
  gettimeofday(&t, NULL);
  return (t.tv_sec - start.tv_sec) * 1000.0 +
    (t.tv_usec - start.tv_usec) / 1000.0;
} // End now()

/* Send heartbeat on the socket descriptor */
void sendHB(hb_pkt_t *pkt) {
  int len = sizeof(hb_pkt_t);
  int res = sendto(sock_fd, (void *)(pkt), len, 0,
		   (struct sockaddr *)&to_addr, sizeof(to_addr));
  if(res != len) {
    //perror("sendHB: sendto");
    return;
  }
  //fprintf(stdout, "--> %ld\n", ntohl(pkt->seqno));
} // End sendHB()

long num_recd = 0;
long num_sent = 0;
double prev_start_time = 0;
long num_lost = 0;
long num_dup = 0;
long next_seqno_expected = 0;

/* Try to receive heartbeat on the listen address */
void tryReceive() {
  fd_set rfds;
  FD_ZERO(&rfds);
  FD_SET(sock_fd, &rfds);
  select(sock_fd+1, &rfds, NULL, NULL, NULL);

  /* Check if there is some data to be read */
  if(FD_ISSET(sock_fd, &rfds)) {
    int len = sizeof(hb_pkt_t);
    hb_pkt_t pkt;
    unsigned int length = sizeof(listen_addr);
    int res = recvfrom(sock_fd, (void *)(&pkt), len, 0,
		       (struct sockaddr*)&listen_addr, &length);
    if(res != len) {
      fprintf(stderr, "tryReceive recvfrom failed, res=%d\n", res);
      perror("recvfrom");
      return;
    }
    num_recd++;
    file_entry_t fe; fe.seqno_recd = ntohl(pkt.seqno); fe.time_recd = now();
    if(fe.seqno_recd > next_seqno_expected) {
      num_lost += fe.seqno_recd - next_seqno_expected;
      next_seqno_expected = fe.seqno_recd+1;
    } else if(fe.seqno_recd < next_seqno_expected) {
      num_dup += next_seqno_expected - fe.seqno_recd;
    } else {
      next_seqno_expected = fe.seqno_recd+1;
    }
    //fprintf(stdout, "<-- %ld\n", fe.seqno_recd);
    //fprintf(stdout, "%ld\t%0.3f\n", fe.seqno_recd, fe.time_recd);
    //fwrite((void*)(&fe), sizeof(file_entry_t), 1, fout);
  }

} // End tryReceive()

/* Print usage of program */
void usage(char *progname) {
  fprintf(stderr, "%s <-s/-r> <remote-host> <hb-period-in-ms> <broadcast-flag(0/1)>\n",
	  progname);
} // End usage()

void print_stats() {
  double now_time = now();
  double time_taken = now_time-prev_start_time;
  double num_expected = time_taken/hbInt;
  double throughput = num_recd*8.0*sizeof(hb_pkt_t)/time_taken;
  throughput /= 1000.0; // In Mbps

  for(int i = 0; i < 60; i++) { fputc('-', stderr); }
  fprintf(stderr, "\n");
  fprintf(stderr, "Now time\t= %0.6f\n", now_time);
  fprintf(stderr, "Num pkts recd\t= %ld\n", num_recd);
  fprintf(stderr, "Num pkts sent\t= %ld\n", num_sent);
  fprintf(stderr, "Num pkts lost\t= %ld\n", num_lost);
  fprintf(stderr, "Num pkts dup\t= %ld\n", num_dup);
  fprintf(stderr, "Time taken\t= %0.4f ms\n", time_taken);
  fprintf(stderr, "Num expected\t= %0.1f\n", num_expected);
  fprintf(stderr, "Expected - Recd\t= %0.1f\n", num_expected-num_recd);
  fprintf(stderr, "Throughput\t= %0.6f Mbps\n", throughput);

  num_sent = num_recd = 0;
  num_lost = num_dup = 0;
  prev_start_time = now_time;
} // End print_stats()

double prev_intr_time = 0;

void intr_catcher(int) {
  print_stats();
  //fprintf(stderr, "Caught Ctrl-C...\n");
  double this_intr_time = now();
  if((this_intr_time-prev_intr_time) <= 1000) {
    fprintf(stderr, "Caught two Ctrl-C within 1sec, exiting...\n");
    fclose(fout);
    exit(0);
  }
  prev_intr_time = this_intr_time;
} // End intr_catcher()

void alarm_catcher(int) {
  print_stats();
  alarm(ALARM_TIME);
} // End alarm_catcher()

#define SEND_MODE 1
#define RECV_MODE 2

int main(int argc, char *argv[]) {
	
   int isbroadcast=0,rc,opt;

  	/* Get command-line arguments */
  if(argc != 5) { usage(argv[0]); exit(11); }
  char mode_str[256];
  strcpy(mode_str, argv[1]);
  strcpy(to_host, argv[2]);
  hbInt = atof(argv[3]);
  isbroadcast=atoi(argv[4]);
  int mode = 0;
  if(strcmp(mode_str, "-s") == 0) { mode = SEND_MODE; }
  else if(strcmp(mode_str, "-r") == 0) { mode = RECV_MODE; }
  else { usage(argv[0]); exit(2); }

  to_port = PORT;
  listen_port = PORT;
  strcpy(foutStr, "/dev/null");

  assert(hbInt > 0);
  if(hbInt <= 50) {
    fprintf(stderr, "Warning: Heartbeat period < 50ms\n");
  }

  /* socket */
  sock_fd = socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(sock_fd < 0) { perror("socket failed"); exit(2); }
  fprintf(stderr, "Socket successful\n");
  /*set socket option for broadcast*/
  if(isbroadcast==1){
	rc=setsockopt(sock_fd,SOL_SOCKET,SO_BROADCAST,(void*)&opt,sizeof(opt));
 	if(rc<0)
		printf("Can not set socket option\n");
  }	
  /* bind */
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(listen_port);

  if(bind(sock_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", listen_port);
    perror("bind failed"); exit(3);
  }
  fprintf(stderr, "Bind successful\n");

  /** open output file */
  fout = fopen(foutStr, "w");
  if(fout == NULL) { perror("Can't open output file"); exit(46); }

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);
  signal(SIGTERM, intr_catcher);

  alarm(ALARM_TIME);
  signal(SIGALRM, alarm_catcher);

  /* init to_addr */
  struct hostent *server = gethostbyname(to_host);
  bzero((char *)&to_addr, sizeof(to_addr));
  bcopy((char *)server->h_addr, (char *)&to_addr.sin_addr.s_addr,
	server->h_length);
  to_addr.sin_family = PF_INET;
  to_addr.sin_port = htons(to_port);

  /* Forever, keep sending and receiving heartbeats */
  gettimeofday(&start, NULL);
  double startTime = 1.0*start.tv_sec + 1e-6 * start.tv_usec;
  num_recd = 0;
  prev_start_time = 0;
  fprintf(stderr, "tv_sec=%ld\n", start.tv_sec);
  fprintf(stderr, "tv_usec=%ld\n", start.tv_usec);
  fprintf(stderr, "Start = %f\n", startTime);
  // put dummy entries into output file
  fprintf(stderr, "sizeof(file_entry_t)=%d\n", sizeof(file_entry_t));
  file_entry_t fe; fe.seqno_recd = -1; fe.time_recd = start.tv_sec;
  //fwrite((void*)(&fe), sizeof(file_entry_t), 1, fout);
  file_entry_t fe1; fe1.seqno_recd = -2; fe1.time_recd = start.tv_usec;
  //fwrite((void*)(&fe1), sizeof(file_entry_t), 1, fout);

  long seqno = 0;
  hb_pkt_t pkt;
  for(;;) {
    double now_time = now();
    // Check if time to send next heartbeat
    if((mode == SEND_MODE) && (now_time >= (seqno*hbInt))) {
      pkt.seqno = htonl(seqno);
      sendHB(&pkt);
      seqno++;
      num_sent++;
    }
    // Try receiving until the time when the next HB has to be sent
    if(mode == RECV_MODE) { tryReceive(); }
  } // End for(;;)

} // End main()
